"""
Michael duPont - michael@mdupont.com
AVWX-Engine : avwx/__init__.py

Contains the primary report classes of avwx: Metar and Taf
"""

# stdlib
import json
from datetime import datetime
from os import path
# module
from avwx import metar, taf, translate, summary, speech, service
from avwx.core import valid_station
from avwx.exceptions import BadStation

INFO_KEYS = ['ICAO', 'Country', 'State', 'City', 'Name', 'IATA',
             'Elevation', 'Latitude', 'Longitude', 'Priority']
INFO_PATH = path.dirname(path.realpath(__file__)) + '/stations.json'
STATIONS = json.load(open(INFO_PATH))


class Report:
    """Base report to take care of station info"""

    def __init__(self, station: str, lang: str = None):
        valid_station(station)
        self.service = service.get_service(station)(self.__class__.__name__.lower())
        self.station = station
        self.lang = lang
        self.last_updated = None
        self.raw = None
        self.data = None
        self.translations = None
        self._station_info = None

    @property
    def station_info(self):
        """Provide basic station info with the keys below"""
        if self._station_info is None:
            if not self.station in STATIONS:
                raise BadStation('Could not find station in the info dict. Check avwx.STATIONS')
            info = [self.station] + STATIONS[self.station]
            self._station_info = dict(zip(INFO_KEYS, info))
        return self._station_info

    def update(self, report: str = None) -> bool:
        """Updates report elements. Not implemented"""
        raise NotImplementedError()


class Metar(Report):
    """Class to handle METAR report data"""

    def update(self, report: str = None) -> bool:
        """Updates raw, data, and translations by fetching and parsing the METAR report"""
        if report is not None:
            self.raw = report
        else:
            raw = self.service.fetch(self.station)
            if raw == self.raw:
                return False
            self.raw = raw
        self.data = metar.parse(self.station, self.raw)
        self.translations = translate.metar(self.data)
        self.last_updated = datetime.utcnow()
        return True

    @property
    def summary(self):
        """Condensed report summary created from translations"""
        if not self.translations:
            self.update()
        return summary.metar(self.translations)

    @property
    def speech(self):
        """Report summary designed to be read by a text-to-speech program"""
        if not self.data:
            self.update()
        return speech.metar(self.data)


class Taf(Report):
    """Class to handle TAF report data"""

    def update(self, report: str = None) -> bool:
        """Updates raw, data, and translations by fetching and parsing the TAF report"""
        if report is not None:
            self.raw = report
        else:
            raw = self.service.fetch(self.station)
            if raw == self.raw:
                return False
            self.raw = raw
        self.data = taf.parse(self.station, self.raw)
        self.translations = translate.taf(self.data)
        self.last_updated = datetime.utcnow()

    @property
    def summary(self):
        """Condensed summary for each forecast created from translations"""
        if not self.translations:
            self.update()
        return [summary.taf(trans) for trans in self.translations['Forecast']]



"""
Michael duPont - michael@mdupont.com
AVWX-Engine : avwx/summary.py

Contains functions for combining translations into a summary string
"""


def metar(trans: {str: object}) -> str:
    """Condense the translation strings into a single report summary string"""
    summary = []
    if 'Wind' in trans and trans['Wind']:
        summary.append('Winds ' + trans['Wind'])
    if 'Visibility' in trans and trans['Visibility']:
        summary.append('Vis ' + trans['Visibility'][:trans['Visibility'].find(' (')].lower())
    if 'Temperature' in trans and trans['Temperature']:
        summary.append('Temp ' + trans['Temperature'][:trans['Temperature'].find(' (')])
    if 'Dewpoint' in trans and trans['Dewpoint']:
        summary.append('Dew ' + trans['Dewpoint'][:trans['Dewpoint'].find(' (')])
    if 'Altimeter' in trans and trans['Altimeter']:
        summary.append('Alt ' + trans['Altimeter'][:trans['Altimeter'].find(' (')])
    if 'Other' in trans and trans['Other']:
        summary.append(trans['Other'])
    if 'Clouds' in trans and trans['Clouds']:
        summary.append(trans['Clouds'].replace(' - Reported AGL', ''))
    return ', '.join(summary)


def taf(trans: {str: object}) -> str:
    """Condense the translation strings into a single forecast summary string"""
    summary = []
    if 'Wind' in trans and trans['Wind']:
        summary.append('Winds ' + trans['Wind'])
    if 'Visibility' in trans and trans['Visibility']:
        summary.append('Vis ' + trans['Visibility'][:trans['Visibility'].find(' (')].lower())
    if 'Altimeter' in trans and trans['Altimeter']:
        summary.append('Alt ' + trans['Altimeter'][:trans['Altimeter'].find(' (')])
    if 'Other' in trans and trans['Other']:
        summary.append(trans['Other'])
    if 'Clouds' in trans and trans['Clouds']:
        summary.append(trans['Clouds'].replace(' - Reported AGL', ''))
    if 'Wind-Shear' in trans and trans['Wind-Shear']:
        summary.append(trans['Wind-Shear'])
    if 'Turbulance' in trans and trans['Turbulance']:
        summary.append(trans['Turbulance'])
    if 'Icing' in trans and trans['Icing']:
        summary.append(trans['Icing'])
    return ', '.join(summary)



"""
Michael duPont - michael@mdupont.com
AVWX-Engine : avwx/taf.py

Contains TAF-specific functions for fetching and parsing
"""

# stdlib
from copy import copy
# module
from avwx import core, service
from avwx.static import NA_UNITS, IN_UNITS, TAF_RMK, TAF_NEWLINE


def fetch(station: str) -> str:
    """Get TAF report for 'station' from www.aviationweather.gov
    Returns TAF report string or raises an error
    Maintains backwards compatability but uses the new AddsRequest object
    """
    return service.get_service(station)('taf').fetch(station)


def parse(station: str, txt: str, delim: str = '<br/>&nbsp;&nbsp;') -> {str: object}:
    """Returns a dictionary of parsed TAF data
    'delim' is the divider between forecast lines. Ex: aviationweather.gov uses '<br/>&nbsp;&nbsp;'
    Keys: Station, Time, Forecast, Remarks, Min-Temp, Max-Temp, Raw-Report, Units
    Oceania stations also have the following keys: Temp-List, Alt-List
    Forecast is list of report dicts in order of time with the following keys:
    Type, Start-Time, End-Time, Flight-Rules, Wind-Direction, Wind-Speed, Wind-Gust, Wind-Shear,
    Visibility, Altimeter, Cloud-List, Icing-List, Turb-List, Other-List, Probability, Raw-Line
    Units is dict of identified units of measurement for each field
    """
    core.valid_station(station)
    retwx = {}
    retwx['Raw-Report'] = txt
    while len(txt) > 3 and txt[:4] in ['TAF ', 'AMD ', 'COR ']:
        txt = txt[4:]
    _, retwx['Station'], retwx['Time'] = core.get_station_and_time(txt[:20].split(' '))
    txt = txt.replace(retwx['Station'], '')
    txt = txt.replace(retwx['Time'], '')
    if core.uses_na_format(retwx['Station']):
        is_international = False
        units = copy(NA_UNITS)
    else:
        is_international = True
        units = copy(IN_UNITS)
    retwx['Remarks'] = ''
    parsed_lines = []
    prob = ''
    lines = txt.strip(' ').split(delim)
    while lines:
        line = lines[0].strip(' ')
        line = core.sanitize_line(line)
        #Remove Remarks from line
        index = core.find_first_in_list(line, TAF_RMK)
        if index != -1:
            retwx['Remarks'] = line[index:]
            line = line[:index].strip(' ')
        #Separate new lines fixed by sanitizeLine
        index = core.find_first_in_list(line, TAF_NEWLINE)
        if index != -1:
            lines.insert(1, line[index + 1:])
            line = line[:index]
        # Remove prob from the beginning of a line
        if line.startswith('PROB'):
            # Add standalone prob to next line
            if len(line) == 6:
                prob = line
                line = ''
            # Add to current line
            elif len(line) > 6:
                prob = line[:6]
                line = line[6:].strip()
        if line:
            # Separate full prob forecast into its own line
            if ' PROB' in line:
                probindex = line.index(' PROB')
                lines.insert(1, line[probindex + 1:])
                line = line[:probindex]
            raw_line = prob + ' ' + line if prob else line
            parsed_line, units = parse_in_line(line, units) if is_international \
                else parse_na_line(line, units)
            parsed_line['Probability'] = prob[4:]
            parsed_line['Raw-Line'] = raw_line
            prob = ''
            parsed_lines.append(parsed_line)
        lines.pop(0)
    if parsed_lines:
        parsed_lines[len(parsed_lines) - 1]['Other-List'], retwx['Max-Temp'], retwx['Min-Temp'] \
            = core.get_temp_min_and_max(parsed_lines[len(parsed_lines) - 1]['Other-List'])
        if not (retwx['Max-Temp'] or retwx['Min-Temp']):
            parsed_lines[0]['Other-List'], retwx['Max-Temp'], retwx['Min-Temp'] \
                = core.get_temp_min_and_max(parsed_lines[0]['Other-List'])
        parsed_lines = core.find_missing_taf_times(parsed_lines)
        parsed_lines = core.get_taf_flight_rules(parsed_lines)
    else:
        retwx['Min-Temp'] = ['', '']
        retwx['Max-Temp'] = ['', '']
    if retwx['Station'][0] == 'A':
        parsed_lines[len(parsed_lines) - 1]['Other-List'], retwx['Alt-List'], retwx['Temp-List'] \
            = core.get_oceania_temp_and_alt(parsed_lines[len(parsed_lines) - 1]['Other-List'])
    retwx['Forecast'] = parsed_lines
    retwx['Units'] = units
    return retwx


def parse_na_line(txt: str, units: {str: str}) -> ({str: object}, {str: str}):
    """Parser for the North American TAF forcast varient"""
    retwx = {}
    wxdata = txt.split(' ')
    wxdata, _, retwx['Wind-Shear'] = core.sanitize_report_list(wxdata, remove_clr_and_skc=False)
    wxdata, retwx['Type'], retwx['Start-Time'], retwx['End-Time'] = core.get_type_and_times(wxdata)
    wxdata, units, retwx['Wind-Direction'], retwx['Wind-Speed'],\
        retwx['Wind-Gust'], _ = core.get_wind(wxdata, units)
    wxdata, units, retwx['Visibility'] = core.get_visibility(wxdata, units)
    wxdata, retwx['Cloud-List'] = core.get_clouds(wxdata)
    retwx['Other-List'], retwx['Altimeter'], retwx['Icing-List'], retwx['Turb-List'] \
        = core.get_taf_alt_ice_turb(wxdata)
    return retwx, units


def parse_in_line(txt: str, units: {str: str}) -> ({str: object}, {str: str}):
    """Parser for the North American TAF forcast varient"""
    retwx = {}
    wxdata = txt.split(' ')
    wxdata, _, retwx['Wind-Shear'] = core.sanitize_report_list(wxdata, remove_clr_and_skc=False)
    wxdata, retwx['Type'], retwx['Start-Time'], retwx['End-Time'] = core.get_type_and_times(wxdata)
    wxdata, units, retwx['Wind-Direction'], retwx['Wind-Speed'],\
        retwx['Wind-Gust'], _ = core.get_wind(wxdata, units)
    if 'CAVOK' in wxdata:
        retwx['Visibility'] = '9999'
        retwx['Cloud-List'] = []
        wxdata.pop(wxdata.index('CAVOK'))
    else:
        wxdata, units, retwx['Visibility'] = core.get_visibility(wxdata, units)
        wxdata, retwx['Cloud-List'] = core.get_clouds(wxdata)
    retwx['Other-List'], retwx['Altimeter'], retwx['Icing-List'], retwx['Turb-List'] \
        = core.get_taf_alt_ice_turb(wxdata)
    return retwx, units



"""
Michael duPont - michael@mdupont.com
AVWX-Engine : avwx/speech.py

Contains functions for converting translations into a speech string
Currently only supports METAR
"""

from avwx import core, translate
from avwx.static import SPOKEN_UNITS, NUMBER_REPL, FRACTIONS


def numbers(num: str) -> str:
    """Returns the spoken version of a number
    Ex: 1.2 -> one point two"""
    if num in FRACTIONS:
        return FRACTIONS[num]
    return ' '.join([NUMBER_REPL[char] for char in num])


def remove_leading_zeros(num: str) -> str:
    """Strips zeros while handling -, M, and empty strings"""
    if not num:
        return num
    if num.startswith('M'):
        ret = 'M' + num[1:].lstrip('0')
    elif num.startswith('-'):
        ret = '-' + num[1:].lstrip('0')
    else:
        ret = num.lstrip('0')
    return ret if ret else '0'


def wind(wdir: str, wspd: str, wgst: str, wvar: [str] = None, unit: str = 'kt') -> str:
    """Format wind details into a spoken word string"""
    if unit in SPOKEN_UNITS:
        unit = SPOKEN_UNITS[unit]
    if wdir not in ('000', 'VRB'):
        wdir = numbers(wdir)
    wvar = wvar if wvar is not None else []
    for i, val in enumerate(wvar):
        wvar[i] = numbers(val)
    return 'Winds ' + translate.wind(wdir, remove_leading_zeros(wspd),
                                     remove_leading_zeros(wgst), wvar,
                                     unit, cardinals=False)


def temperature(header: str, temp: str, unit: str = 'C') -> str:
    """Format temperature details into a spoken word string"""
    if core.is_unknown(temp):
        return header + ' Unknown'
    if unit in SPOKEN_UNITS:
        unit = SPOKEN_UNITS[unit]
    temp = numbers(remove_leading_zeros(temp))
    use_s = '' if temp in ('one', 'minus one') else 's'
    return ' '.join((header, temp, 'degree' + use_s, unit))


def unpack_fraction(num: str) -> str:
    """Returns unpacked fraction string 5/2 -> 2 1/2"""
    nums = [int(n) for n in num.split('/')]
    if nums[0] > nums[1]:
        over = nums[0] // nums[1]
        rem = nums[0] % nums[1]
        return '{} {}/{}'.format(over, rem, nums[1])
    return num


def visibility(vis: str, unit: str = 'm') -> str:
    """Format visibility details into a spoken word string"""
    if core.is_unknown(vis):
        return 'Visibility Unknown'
    elif vis.startswith('M'):
        vis = 'less than ' + numbers(remove_leading_zeros(vis[1:]))
    elif vis.startswith('P'):
        vis = 'greater than ' + numbers(remove_leading_zeros(vis[1:]))
    elif '/' in vis:
        vis = unpack_fraction(vis)
        vis = ' and '.join([numbers(remove_leading_zeros(n)) for n in vis.split(' ')])
    else:
        vis = translate.visibility(vis, unit=unit)
        if unit == 'm':
            unit = 'km'
        vis = vis[:vis.find(' (')].lower().replace(unit, '').strip()
        vis = numbers(remove_leading_zeros(vis))
    ret = 'Visibility ' + vis
    if unit in SPOKEN_UNITS:
        ret += ' ' + SPOKEN_UNITS[unit]
        if not (('one half' in vis and ' and ' not in vis) or 'of a' in vis):
            ret += 's'
    else:
        ret += unit
    return ret


def altimeter(alt: str, unit: str = 'inHg') -> str:
    """Format altimeter details into a spoken word string"""
    ret = 'Altimeter '
    if core.is_unknown(alt):
        ret += 'Unknown'
    elif unit == 'inHg':
        ret += numbers(alt[:2]) + ' point ' + numbers(alt[2:])
    elif unit == 'hPa':
        ret += numbers(alt)
    return ret


def other(wxcodes: [str]) -> str:
    """Format wx codes into a spoken word string"""
    ret = []
    for item in wxcodes:
        item = translate.wxcode(item)
        if item.startswith('Vicinity'):
            item = item.lstrip('Vicinity ') + ' in the Vicinity'
        ret.append(item)
    return '. '.join(ret)


def metar(wxdata: {str: object}) -> str:
    """Convert wxdata into a string for text-to-speech"""
    speech = []
    units = wxdata['Units']
    if wxdata['Wind-Direction'] and wxdata['Wind-Speed']:
        speech.append(wind(wxdata['Wind-Direction'], wxdata['Wind-Speed'],
                           wxdata['Wind-Gust'], wxdata['Wind-Variable-Dir'],
                           units['Wind-Speed']))
    if wxdata['Visibility']:
        speech.append(visibility(wxdata['Visibility'], units['Visibility']))
    if wxdata['Temperature']:
        speech.append(temperature('Temperature', wxdata['Temperature'], units['Temperature']))
    if wxdata['Dewpoint']:
        speech.append(temperature('Dew point', wxdata['Dewpoint'], units['Temperature']))
    if wxdata['Altimeter']:
        speech.append(altimeter(wxdata['Altimeter'], units['Altimeter']))
    if wxdata['Other-List']:
        speech.append(other(wxdata['Other-List']))
    speech.append(translate.clouds(wxdata['Cloud-List'],
                                   units['Altitude']).replace(' - Reported AGL', ''))
    return ('. '.join([l for l in speech if l])).replace(',', '.')



"""
Michael duPont - michael@mdupont.com
AVWX-Engine : avwx/static.py

Contains static objects for internal and external use
"""

# Station Location Identifiers (first character)
NA_REGIONS = ['C', 'K', 'P', 'T']
IN_REGIONS = ['A', 'B', 'D', 'E', 'F', 'G', 'H', 'L', 'N', 'O', 'R', 'S', 'U', 'V', 'W', 'Y', 'Z']
# The Central American region is split. Therefore we need to use the first two letters
M_NA_REGIONS = ['MB', 'MM', 'MT', 'MY']
M_IN_REGIONS = ['MD', 'MG', 'MH', 'MK', 'MN', 'MP', 'MR', 'MS', 'MU', 'MW', 'MZ']

# North American variant units
NA_UNITS = {
    'Wind-Speed': 'kt',
    'Visibility': 'sm',
    'Altitude': 'ft',
    'Temperature': 'C',
    'Altimeter': 'inHg'
}
# International variant units
IN_UNITS = {
    'Wind-Speed': 'kt',
    'Visibility': 'm',
    'Altitude': 'ft',
    'Temperature': 'C',
    'Altimeter': 'hPa'
}

# List of flight rules abreviations
FLIGHT_RULES = ['VFR', 'MVFR', 'IFR', 'LIFR']

# List of cloud layer abreviations
CLOUD_LIST = ['FEW', 'SCT', 'BKN', 'OVC']

# Dictionary associating WX codes with descriptions
WX_TRANSLATIONS = {
    'BC': 'Patchy',
    'BL': 'Blowing',
    'BR': 'Mist',
    'DS': 'Duststorm',
    'DU': 'Wide Dust',
    'DZ': 'Drizzle',
    'FC': 'Funnel Cloud',
    'FG': 'Fog',
    'FU': 'Smoke',
    'FZ': 'Freezing',
    'GR': 'Hail',
    'GS': 'Small Hail',
    'HZ': 'Haze',
    'IC': 'Ice Crystals',
    'MI': 'Shallow',
    'PL': 'Ice Pellets',
    'PO': 'Dust Whirls',
    'PR': 'Partial',
    'RA': 'Rain',
    'SA': 'Sand',
    'SG': 'Snow Grains',
    'SH': 'Showers',
    'SN': 'Snow',
    'SQ': 'Squall',
    'SS': 'Sandstorm',
    'SY': 'Spray',
    'TS': 'Thunderstorm',
    'UP': 'Unknown Precip',
    'VA': 'Volcanic Ash',
    'VC': 'Vicinity'
}

# Dictionary associating cloud layer and cloud codes with descriptions
CLOUD_TRANSLATIONS = {
    'OVC': 'Overcast layer at {0}{1}',
    'BKN': 'Broken layer at {0}{1}',
    'SCT': 'Scattered clouds at {0}{1}',
    'FEW': 'Few clouds at {0}{1}',
    'VV': 'Vertical visibility up to {0}{1}',
    'CLR': 'Sky Clear',
    'SKC': 'Sky Clear',
    'AC': 'Altocumulus',
    'ACC': 'Altocumulus Castellanus',
    'AS': 'Altostratus',
    'CB': 'Cumulonimbus',
    'CC': 'Cirrocumulus',
    'CI': 'Cirrus',
    'CS': 'Cirrostratus',
    'CU': 'Cumulus',
    'FC': 'Fractocumulus',
    'FS': 'Fractostratus',
    'NS': 'Nimbostratus',
    'SC': 'Stratocumulus',
    'ST': 'Stratus',
    'TCU': 'Towering Cumulus'
}

# Dictionary associating turbulance report IDs with descriptions
TURBULANCE_CONDITIONS = {
    '0': 'None',
    '1': 'Light turbulence',
    '2': 'Occasional moderate turbulence in clear air',
    '3': 'Frequent moderate turbulence in clear air',
    '4': 'Occasional moderate turbulence in clouds',
    '5': 'Frequent moderate turbulence in clouds',
    '6': 'Occasional severe turbulence in clear air',
    '7': 'Frequent severe turbulence in clear air',
    '8': 'Occasional severe turbulence in clouds',
    '9': 'Frequent severe turbulence in clouds',
    'X': 'Extreme turbulence'
}

# Dictionary associating icing report IDs with descriptions
ICING_CONDITIONS = {
    '0': 'No icing',
    '1': 'Light icing',
    '2': 'Light icing in clouds',
    '3': 'Light icing in precipitation',
    '4': 'Moderate icing',
    '5': 'Moderate icing in clouds',
    '6': 'Moderate icing in precipitation',
    '7': 'Severe icing',
    '8': 'Severe icing in clouds',
    '9': 'Severe icing in precipitation'
}

# Strings signifying the start of the remarks section of a new TAF time period
METAR_RMK = [' BLU', ' BLU+', ' WHT', ' GRN', ' YLO', ' AMB', ' RED', ' BECMG', ' TEMPO',
             ' INTER', ' NOSIG', ' RMK', ' WIND', ' QFE', ' QFF', ' INFO', ' RWY', ' CHECK']
TAF_RMK = ['RMK ', 'AUTOMATED ', 'COR ', 'AMD ', 'LAST ', 'FCST ',
           'CANCEL ', 'CHECK ', 'WND ', 'MOD ', ' BY', ' QFE', ' QFF']
TAF_NEWLINE = [' INTER ', ' FM', ' BECMG ', ' TEMPO ']

# Units required to be translated in order to be spoken properly
SPOKEN_UNITS = {
    'sm': 'mile',
    'km': 'kilometer',
    'C': 'Celsius',
    'F': 'Fahrenheit'
}

# Dictionary associating algebraic signs with their spoken version
NUMBER_REPL = {
    '.': 'point',
    '-': 'minus',
    'M': 'minus',
    '0': 'zero',
    '1': 'one',
    '2': 'two',
    '3': 'three',
    '4': 'four',
    '5': 'five',
    '6': 'six',
    '7': 'seven',
    '8': 'eight',
    '9': 'nine'
}

# Dictionary associating fraction strings with their spoken version
FRACTIONS = {
    '1/4': 'one quarter of a',
    '1/2': 'one half',
    '3/4': 'three quarters of a'
}

# Static remarks translation elements
REMARKS_ELEMENTS = {
    '$': 'ASOS requires maintenance',
    'AO1': 'Automated with no precipitation sensor',
    'AO2': 'Automated with precipitation sensor',
    'BINOVC': 'Breaks in Overcast',
    'FZRANO': 'Freezing rain information not available',
    'NOSPECI': 'No SPECI reports taken',
    'P0000': 'Trace amount of rain in the last hour',
    'PNO': 'Precipitation amount not available',
    'PRESFR': 'Pressure Falling Rapidly',
    'PRESRR': 'Pressure Rising Rapidly',
    'PWINO': 'Precipitation identifier information not available',
    'RVRNO': 'RVR missing',
    'TSNO': 'Thunderstorm information not available'
}

# Static remarks translation groups
REMARKS_GROUPS = {
    ' ACFT MSHP ': 'Aircraft mishap'
}

PRESSURE_TENDENCIES = {
    '0': 'Increasing, then decreasing',
    '1': 'Increasing, then steady',
    '2': 'Increasing steadily or unsteadily',
    '3': 'Decreasing or steady, then increasing',
    '4': 'Steady',
    '5': 'Decreasing, then increasing',
    '6': 'Decreasing, then steady',
    '7': 'Decreasing steadily or unsteadily',
    '8': 'Steady or increasing, then decreasing',
}



"""
Michael duPont - michael@mdupont.com
AVWX-Engine : avwx/service.py

Classes for retrieving raw report strings
"""

# library
import requests
from xmltodict import parse as parsexml
# module
from avwx.core import valid_station
from avwx.exceptions import InvalidRequest

class Service(object):
    """Base Service class for fetching reports"""

    # Service URL must accept report type and station via .format()
    url: str = None

    def __init__(self, request_type: str):
        self.rtype = request_type

    def make_err(self, body: str, key: str = 'report path') -> InvalidRequest:
        """Returns an InvalidRequest exception with formatted error message"""
        msg = f'Could not find {key} in {self.__class__.__name__} response\n'
        return InvalidRequest(msg + body)

    def _extract(self, raw: str) -> str:
        """Extracts report from response. Not implemented"""
        raise NotImplementedError()

    def fetch(self, station: str) -> str:
        """Fetches a report string from the service"""
        valid_station(station)
        resp = requests.get(self.url.format(self.rtype, station)).text
        return self._extract(resp)


class NOAA(Service):
    """Requests data from NOAA ADDS"""

    url = (
        'https://aviationweather.gov/adds/dataserver_current/httpparam'
        '?dataSource={0}s'
        '&requestType=retrieve'
        '&format=XML'
        '&stationString={1}'
        '&hoursBeforeNow=2'
    )

    def _extract(self, raw: str) -> str:
        """Extracts the raw_report element from XML response"""
        resp = parsexml(raw)
        try:
            report = resp['response']['data'][self.rtype.upper()]
        except:
            raise self.make_err(raw)
        if isinstance(report, dict):
            return report['raw_text']
        elif isinstance(report, list) and report:
            return report[0]['raw_text']
        else:
            raise self.make_err(raw, '"raw_text"')


class AMO(Service):
    """Requests data from AMO KMA for Korean stations"""

    url = 'http://amoapi.kma.go.kr/amoApi/{0}?icao={1}'

    def _extract(self, raw: str) -> str:
        """Extracts the report message from XML response"""
        resp = parsexml(raw)
        try:
            report = resp['response']['body']['items']['item'][self.rtype.lower() + 'Msg']
        except:
            raise self.make_err(raw)
        # Replace line breaks
        report = report.replace('\n', '')
        # Remove excess leading and trailing data
        report = report.lstrip(self.rtype.upper()).rstrip('=')
        # Make every element single-spaced and stripped
        return ' '.join(report.split())


PREFERRED = {
    'RK': AMO
}


def get_service(station: str) -> Service:
    """Returns the preferred service for a given station if available
    Defaults to NOAA
    """
    for prefix in PREFERRED:
        if station.startswith(prefix):
            return PREFERRED[prefix]
    return NOAA



"""
Michael duPont - michael@mdupont.com
AVWX-Engine : avwx/core.py

Contains the core parsing and indent functions of avwx
"""

# stdlib
from copy import copy
from itertools import permutations
# module
from avwx.exceptions import BadStation
from avwx.static import CLOUD_LIST, CLOUD_TRANSLATIONS, METAR_RMK, \
    NA_REGIONS, IN_REGIONS, M_NA_REGIONS, M_IN_REGIONS, FLIGHT_RULES


def valid_station(station: str):
    """Checks the validity of station ident and aises BadStation exception if needed
    This function doesn't return anything. It merely raises a BadStation error if needed
    """
    station = station.strip()
    if len(station) != 4:
        raise BadStation('ICAO station idents must be four characters long')
    uses_na_format(station)


def uses_na_format(station: str) -> bool:
    """Returns True if the station uses the North American format,
    False if the International format
    """
    if station[0] in NA_REGIONS:
        return True
    elif station[0] in IN_REGIONS:
        return False
    elif station[:2] in M_NA_REGIONS:
        return True
    elif station[:2] in M_IN_REGIONS:
        return False
    raise BadStation("Station doesn't start with a recognized character set")


def is_unknown(val: str) -> bool:
    """Returns True if val contains only '/' characters"""
    return val == '/' * len(val)


def find_first_in_list(txt: str, str_list: [str]) -> int:
    """Returns the index of the earliest occurence of an item from a list in a string
    Ex: find_first_in_list('foobar', ['bar', 'fin']) -> 3
    """
    start = len(txt) + 1
    for item in str_list:
        if start > txt.find(item) > -1:
            start = txt.find(item)
    return start if len(txt) + 1 > start > -1 else -1


def get_remarks(txt) -> ([str], str):
    """Returns the report split into components and the remarks string
    Remarks can include items like RMK and on, NOSIG and on, and BECMG and on
    """
    txt = txt.replace('?', '').strip(' ')
    # First look for Altimeter in txt
    alt_index = len(txt) + 1
    for item in [' A2', ' A3', ' Q1', ' Q0', ' Q9']:
        index = txt.find(item)
        if len(txt) - 6 > index > -1 and txt[index + 2:index + 6].isdigit():
            alt_index = index
    # Then look for earliest remarks 'signifier'
    sig_index = find_first_in_list(txt, METAR_RMK)
    if sig_index == -1:
        sig_index = len(txt) + 1
    if sig_index > alt_index > -1:
        return txt[:alt_index + 6].strip().split(' '), txt[alt_index + 7:]
    elif alt_index > sig_index > -1:
        return txt[:sig_index].strip().split(' '), txt[sig_index + 1:]
    return txt.strip().split(' '), ''


STR_REPL = {' C A V O K ': ' CAVOK ', '?': ' '}


def sanitize_report_string(txt: str) -> str:
    """Provides sanitization for operations that work better when the report is a string
    Returns the first pass sanitized report string
    """
    if len(txt) < 4:
        return txt
    # Prevent changes to station ID
    stid, txt = txt[:4], txt[4:]
    # Replace invalid key-value pairs
    for key, rep in STR_REPL.items():
        txt = txt.replace(key, rep)
    # Check for missing spaces in front of cloud layers
    # Ex: TSFEW004SCT012FEW///CBBKN080
    for cloud in CLOUD_LIST:
        if cloud in txt and ' ' + cloud not in txt:
            start, counter = 0, 0
            while txt.count(cloud) != txt.count(' ' + cloud):
                cloud_index = start + txt[start:].find(cloud)
                if len(txt[cloud_index:]) >= 3:
                    target = txt[cloud_index + len(cloud):cloud_index + len(cloud) + 3]
                    if target.isdigit() or not target.strip('/'):
                        txt = txt[:cloud_index] + ' ' + txt[cloud_index:]
                start = cloud_index + len(cloud) + 1
                # Prevent infinite loops
                if counter > txt.count(cloud):
                    break
                counter += 1
    return stid + txt


LINE_FIXES = {'TEMP0': 'TEMPO', 'TEMP O': 'TEMPO', 'TMPO': 'TEMPO', 'TE MPO': 'TEMPO',
              'TEMP ': 'TEMPO ', ' EMPO': ' TEMPO', 'TEMO': 'TEMPO', 'T EMPO': 'TEMPO',
              'BECM G': 'BECMG', 'BEMCG': 'BECMG', 'BE CMG': 'BECMG', 'BEMG': 'BECMG',
              ' BEC ': ' BECMG ', 'BCEMG': 'BECMG', 'B ECMG': 'BECMG'}


def sanitize_line(txt: str) -> str:
    """Fixes common mistakes with 'new line' signifiers so that they can be recognized"""
    for key in LINE_FIXES:
        index = txt.find(key)
        if index > -1:
            txt = txt[:index] + LINE_FIXES[key] + txt[index + len(key):]
    #Fix when space is missing following new line signifiers
    for item in ['BECMG', 'TEMPO']:
        if item in txt and item + ' ' not in txt:
            index = txt.find(item) + len(item)
            txt = txt[:index] + ' ' + txt[index:]
    return txt


def extra_space_exists(str1: str, str2: str) -> bool:
    """Return True if a space shouldn't exist between two items"""
    ls1, ls2 = len(str1), len(str2)
    if str1.isdigit():
        # 10 SM
        if str2 in ['SM', '0SM']:
            return True
        # 12 /10
        if ls2 > 2 and str2[0] == '/' and str2[1:].isdigit():
            return True
    if str2.isdigit():
        # OVC 040
        if str1 in CLOUD_LIST:
            return True
        # 12/ 10
        if ls1 > 2 and str1.endswith('/') and str1[:-1].isdigit():
            return True
        # 12/1 0
        if ls2 == 1 and ls1 > 3 and str1[:2].isdigit() and '/' not in str1 and str1[3:].isdigit():
            return True
        # Q 1001
        if str1 in ['Q', 'A']:
            return True
    # 36010G20 KT
    if str2 == 'KT' and str1[-1].isdigit() \
        and (str1[:5].isdigit() or (str1.startswith('VRB') and str1[3:5].isdigit())):
        return True
    # 36010K T
    if str2 == 'T' and ls1 == 6 \
        and (str1[:5].isdigit() or (str1.startswith('VRB') and str1[3:5].isdigit())) and str1[5] == 'K':
        return True
    # OVC022 CB
    if str2 in CLOUD_TRANSLATIONS and str2 not in CLOUD_LIST and ls1 >= 3 and str1[:3] in CLOUD_LIST:
        return True
    # FM 122400
    if str1 in ['FM', 'TL'] and (str2.isdigit() or (str2.endswith('Z') and str2[:-1].isdigit())):
        return True
    # TX 20/10
    if str1 in ['TX', 'TN'] and str2.find('/') != -1:
        return True
    return False


ITEM_REMV = ['AUTO', 'COR', 'NSC', 'NCD', '$', 'KT', 'M', '.', 'RTD', 'SPECI', 'METAR', 'CORR']
ITEM_REPL = {'CALM': '00000KT'}
VIS_PERMUTATIONS = [''.join(p) for p in permutations('P6SM')]
VIS_PERMUTATIONS.remove('6MPS')


def sanitize_report_list(wxdata: [str], remove_clr_and_skc: bool = True) -> ([str], [str], str):
    """Sanitize wxData
    We can remove and identify "one-off" elements and fix other issues before parsing a line
    We also return the runway visibility and wind shear since they are very easy to recognize
    and their location in the report is non-standard
    """
    shear = ''
    runway_vis = []
    for i, item in reversed(list(enumerate(wxdata))):
        ilen = len(item)
        # Remove elements containing only '/'
        if is_unknown(item):
            wxdata.pop(i)
        # Identify Runway Visibility
        elif ilen > 4 and item[0] == 'R' \
            and (item[3] == '/' or item[4] == '/') and item[1:3].isdigit():
            runway_vis.append(wxdata.pop(i))
        # Remove RE from wx codes, REVCTS -> VCTS
        elif ilen in [4, 6] and item.startswith('RE'):
            wxdata[i] = item[2:]
        # Fix a slew of easily identifiable conditions where a space does not belong
        elif i and extra_space_exists(wxdata[i - 1], item):
            wxdata[i - 1] += wxdata.pop(i)
        # Remove spurious elements
        elif item in ITEM_REMV:
            wxdata.pop(i)
        # Remove 'Sky Clear' from METAR but not TAF
        elif remove_clr_and_skc and item in ['CLR', 'SKC']:
            wxdata.pop(i)
        # Replace certain items
        elif item in ITEM_REPL:
            wxdata[i] = ITEM_REPL[item]
        # Remove ammend signifier from start of report ('CCA', 'CCB',etc)
        elif ilen == 3 and item.startswith('CC') and item[2].isalpha():
            wxdata.pop(i)
        # Identify Wind Shear
        elif ilen > 6 and item.startswith('WS') and item[5] == '/':
            shear = wxdata.pop(i).replace('KT', '')
        # Fix inconsistant 'P6SM' Ex: TP6SM or 6PSM -> P6SM
        elif ilen > 3 and item[-4:] in VIS_PERMUTATIONS:
            wxdata[i] = 'P6SM'
        # Fix wind T
        elif (ilen == 6 and item[5] in ['K', 'T'] and (item[:5].isdigit() or item.startswith('VRB'))) \
            or (ilen == 9 and item[8] in ['K', 'T'] and item[5] == 'G' and (item[:5].isdigit() or item.startswith('VRB'))):
            wxdata[i] = item[:-1] + 'KT'
        # Fix joined TX-TN
        elif ilen > 16 and len(item.split('/')) == 3:
            if item.startswith('TX') and 'TN' not in item:
                tn_index = item.find('TN')
                wxdata.insert(i + 1, item[:tn_index])
                wxdata[i] = item[tn_index:]
            elif item.startswith('TN') and item.find('TX') != -1:
                tx_index = item.find('TX')
                wxdata.insert(i + 1, item[:tx_index])
                wxdata[i] = item[tx_index:]
    return wxdata, runway_vis, shear


def is_not_tempo_or_prob(report_type: str) -> bool:
    """Returns True if report type is TEMPO or PROB__"""
    return report_type != 'TEMPO' and not (len(report_type) == 6 and report_type.startswith('PROB'))


def get_altimeter(wxdata: [str], units: {str: str}, version: str = 'NA') -> ([str], {str: str}, str):
    """Returns the report list and the removed altimeter item
    Version is 'NA' (North American / default) or 'IN' (International)
    """
    if not wxdata:
        return wxdata, units, ''
    altimeter = ''
    if version == 'NA':
        if wxdata[-1][0] == 'A':
            altimeter = wxdata.pop()[1:]
        elif wxdata[-1][0] == 'Q':
            units['Altimeter'] = 'hPa'
            altimeter = wxdata.pop()[1:].lstrip('.')
        elif len(wxdata[-1]) == 4 and wxdata[-1].isdigit():
            altimeter = wxdata.pop()
    elif version == 'IN':
        if wxdata[-1][0] == 'Q':
            altimeter = wxdata.pop()[1:].lstrip('.')
            if altimeter.find('/') != -1:
                altimeter = altimeter[:altimeter.find('/')]
        elif wxdata[-1][0] == 'A':
            units['Altimeter'] = 'inHg'
            altimeter = wxdata.pop()[1:]
    #Some stations report both, but we only need one
    if wxdata and (wxdata[-1][0] == 'A' or wxdata[-1][0] == 'Q'):
        wxdata.pop()
    return wxdata, units, altimeter


def get_taf_alt_ice_turb(wxdata: [str]) -> ([str], str, [str], [str]):
    """Returns the report list and removed: Altimeter string, Icing list, Turbulance list
    """
    altimeter = ''
    icing, turbulence = [], []
    for i, item in reversed(list(enumerate(wxdata))):
        if len(item) > 6 and item.startswith('QNH') and item[3:7].isdigit():
            altimeter = wxdata.pop(i)[3:7]
        elif item.isdigit():
            if item[0] == '6':
                icing.append(wxdata.pop(i))
            elif item[0] == '5':
                turbulence.append(wxdata.pop(i))
    return wxdata, altimeter, icing, turbulence


def is_possible_temp(temp: str) -> bool:
    """Returns True if all characters are digits or 'M' (for minus)"""
    for char in temp:
        if not (char.isdigit() or char == 'M'):
            return False
    return True


def get_temp_and_dew(wxdata: str) -> ([str], str, str):
    """Returns the report list and removed temperature and dewpoint strings"""
    for i, item in reversed(list(enumerate(wxdata))):
        if '/' in item:
            #///07
            if item[0] == '/':
                item = '/' + item.strip('/')
            #07///
            elif item[-1] == '/':
                item = item.strip('/') + '/'
            tempdew = item.split('/')
            if len(tempdew) != 2:
                continue
            valid = True
            for j, temp in enumerate(tempdew):
                if temp in ['MM', 'XX']:
                    tempdew[j] = ''
                elif not is_possible_temp(temp):
                    valid = False
                    break
            if valid:
                wxdata.pop(i)
                return wxdata, tempdew[0], tempdew[1]
    return wxdata, '', ''


def get_station_and_time(wxdata: [str]) -> ([str], str, str):
    """Returns the report list and removed station ident and time strings"""
    station = wxdata.pop(0)
    if wxdata and wxdata[0].endswith('Z') and wxdata[0][:-1].isdigit():
        rtime = wxdata.pop(0)
    elif wxdata and len(wxdata[0]) == 6 and wxdata[0].isdigit():
        rtime = wxdata.pop(0) + 'Z'
    else:
        rtime = ''
    return wxdata, station, rtime


def get_wind(wxdata: [str], units: {str: str}) -> ([str], {str: str}, str, str, str, [str]):
    """Returns the report list and removed:
    Direction string, speed string, gust string, variable direction list"""
    direction, speed, gust = '', '', ''
    variable = []
    if wxdata:
        item = copy(wxdata[0])
        for rep in ['(E)']:
            item = item.replace(rep, '')
        item = item.replace('O', '0')
        #09010KT, 09010G15KT
        if item.endswith('KT') \
            or item.endswith('KTS') \
            or item.endswith('MPS') \
            or item.endswith('KMH') \
            or ((len(item) == 5 or (len(item) >= 8 and item.find('G') != -1) and item.find('/') == -1)
            and (item[:5].isdigit() or (item.startswith('VRB') and item[3:5].isdigit()))):
            #In order of frequency
            if item.endswith('KT'):
                item = item.replace('KT', '')
            elif item.endswith('KTS'):
                item = item.replace('KTS', '')
            elif item.endswith('MPS'):
                units['Wind-Speed'] = 'm/s'
                item = item.replace('MPS', '')
            elif item.endswith('KMH'):
                units['Wind-Speed'] = 'km/h'
                item = item.replace('KMH', '')
            direction = item[:3]
            if 'G' in item:
                g_index = item.find('G')
                gust = item[g_index + 1:]
                speed = item[3:g_index]
            else:
                speed = item[3:]
            wxdata.pop(0)
        # elif len(item) > 5 and item[3] == '/' and item[:3].isdigit() and item[4:6].isdigit():
        #     direction = item[:3]
        #     if item.find('G') != -1:
        #         print('Found second G: {0}'.format(item))
        #         gIndex = item.find('G')
        #         gust = item[gIndex+1:gIndex+3]
        #         speed = item[4:item.find('G')]
        #     else:
        #         speed = item[4:]
        #     wxdata.pop(0)
    #Separated Gust
    if wxdata and 1 < len(wxdata[0]) < 4 and wxdata[0][0] == 'G' and wxdata[0][1:].isdigit():
        gust = wxdata.pop(0)[1:]
    #Variable Wind Direction
    if wxdata and len(wxdata[0]) == 7 and wxdata[0][:3].isdigit() \
        and wxdata[0][3] == 'V' and wxdata[0][4:].isdigit():
        variable = wxdata.pop(0).split('V')
    return wxdata, units, direction, speed, gust, variable


def get_visibility(wxdata: [str], units: {str: str}) -> ([str], {str: str}, str):
    """Returns the report list and removed visibility string"""
    visibility = ''
    if wxdata:
        item = copy(wxdata[0])
        # Vis reported in statue miles
        if item.endswith('SM'):  # 10SM
            if item == 'P6SM':
                visibility = 'P6'
            elif item == 'M1/4SM':
                visibility = 'M1/4'
            elif '/' not in item:
                visibility = str(int(item[:item.find('SM')]))
            else:
                visibility = item[:item.find('SM')]  # 1/2SM
            wxdata.pop(0)
            units['Visibility'] = 'sm'
        # Vis reported in meters
        elif len(item) == 4 and item.isdigit():
            visibility = wxdata.pop(0)
            units['Visibility'] = 'm'
        elif 7 >= len(item) >= 5 and item[:4].isdigit() \
            and (item[4] in ['M', 'N', 'S', 'E', 'W'] or item[4:] == 'NDV'):
            visibility = wxdata.pop(0)[:4]
            units['Visibility'] = 'm'
        elif len(item) == 5 and item[1:5].isdigit() and item[0] in ['M', 'P', 'B']:
            visibility = wxdata.pop(0)[1:5]
            units['Visibility'] = 'm'
        elif item.endswith('KM') and item[:item.find('KM')].isdigit():
            visibility = item[:item.find('KM')] + '000'
            wxdata.pop(0)
            units['Visibility'] = 'm'
        # Vis statute miles but split Ex: 2 1/2SM
        elif len(wxdata) > 1 and wxdata[1].endswith('SM') and '/' in wxdata[1] and item.isdigit():
            vis1 = wxdata.pop(0)  # 2
            vis2 = wxdata.pop(0).replace('SM', '')  # 1/2
            visibility = str(int(vis1) * int(vis2[2]) + int(vis2[0])) + vis2[1:]  # 5/2
            units['Visibility'] = 'sm'
    return wxdata, units, visibility


# TAF line report type and start/end times
def get_type_and_times(wxdata: [str]) -> ([str], str, str, str):
    """Returns the report list and removed:
    Report type string, start time string, end time string
    """
    report_type, start_time, end_time = 'BASE', '', ''
    #TEMPO, BECMG, INTER
    if wxdata and wxdata[0] in ['TEMPO', 'BECMG', 'INTER']:
        report_type = wxdata.pop(0)
    #PROB[30,40]
    elif wxdata and len(wxdata[0]) == 6 and wxdata[0].startswith('PROB'):
        report_type = wxdata.pop(0)
    #1200/1306
    if wxdata and len(wxdata[0]) == 9 and wxdata[0][4] == '/' \
        and wxdata[0][:4].isdigit() and wxdata[0][5:].isdigit():
        start_time, end_time = wxdata.pop(0).split('/')
    #FM120000
    elif wxdata and len(wxdata[0]) > 7 and wxdata[0].startswith('FM'):
        report_type = 'FROM'
        if '/' in wxdata[0] and wxdata[0][2:].split('/')[0].isdigit() \
            and wxdata[0][2:].split('/')[1].isdigit():
            start_time, end_time = wxdata.pop(0)[2:].split('/')
        elif wxdata[0][2:8].isdigit():
            start_time = wxdata.pop(0)[2:6]
        #TL120600
        if wxdata and len(wxdata[0]) > 7 and wxdata[0].startswith('TL') \
            and wxdata[0][2:8].isdigit():
            end_time = wxdata.pop(0)[2:6]
    return wxdata, report_type, start_time, end_time


def find_missing_taf_times(lines: [str]) -> [str]:
    """Fix any missing time issues (except for error/empty lines)"""
    last_fm_line = 0
    for i, line in enumerate(lines):
        if line['End-Time'] == '' and is_not_tempo_or_prob(line['Type']):
            last_fm_line = i
            if i < len(lines) - 1:
                for report in lines[i + 1:]:
                    if is_not_tempo_or_prob(report['Type']):
                        line['End-Time'] = report['Start-Time']
                        break
    #Special case for final forcast
    if last_fm_line > 0:
        lines[last_fm_line]['End-Time'] = lines[0]['End-Time']
    return lines


def get_temp_min_and_max(wxlist: [str]) -> ([str], str, str):
    """Pull out Max temp at time and Min temp at time items from wx list
    """
    temp_max, temp_min = '', ''
    for i, item in reversed(list(enumerate(wxlist))):
        if len(item) > 6 and item[0] == 'T' and '/' in item:
            # TX12/1316Z
            if item[1] == 'X':
                temp_max = wxlist.pop(i)
            # TNM03/1404Z
            elif item[1] == 'N':
                temp_min = wxlist.pop(i)
            # TM03/1404Z T12/1316Z -> Will fix TN/TX
            elif item[1] == 'M' or item[1].isdigit():
                if temp_min:
                    if int(temp_min[2:temp_min.find('/')].replace('M', '-')) \
                        > int(item[1:item.find('/')].replace('M', '-')):
                        temp_max = 'TX' + temp_min[2:]
                        temp_min = 'TN' + item[1:]
                    else:
                        temp_max = 'TX' + item[1:]
                else:
                    temp_min = 'TN' + item[1:]
                wxlist.pop(i)
    return wxlist, temp_max, temp_min


def get_digit_list(alist: [str], from_index: int) -> ([str], [str]):
    """Returns a list of items removed from a given list of strings
    that are all digits from 'from_index' until hitting a non-digit item
    """
    ret = []
    alist.pop(from_index)
    while len(alist) > from_index and alist[from_index].isdigit():
        ret.append(alist.pop(from_index))
    return alist, ret


def get_oceania_temp_and_alt(wxlist: [str]) -> ([str], [str], [str]):
    """Get Temperature and Altimeter lists for Oceania TAFs"""
    tlist, qlist = [], []
    if 'T' in wxlist:
        wxlist, tlist = get_digit_list(wxlist, wxlist.index('T'))
    if 'Q' in wxlist:
        wxlist, qlist = get_digit_list(wxlist, wxlist.index('Q'))
    return wxlist, tlist, qlist


def sanitize_cloud(cloud: str) -> str:
    """Fix rare cloud layer issues"""
    if len(cloud) < 4:
        return cloud
    if not cloud[3].isdigit() and cloud[3] != '/':
        if cloud[3] == 'O':
            cloud = cloud[:3] + '0' + cloud[4:]  # Bad "O": FEWO03 -> FEW003
        else:  # Move modifiers to end: BKNC015 -> BKN015C
            cloud = cloud[:3] + cloud[4:] + cloud[3]
    return cloud


def split_cloud(cloud: str, begins_with_vv: bool) -> [str]:
    """Transforms a cloud string into a list of strings: [Type, Height (, Optional Modifier)]"""
    split = []
    cloud = sanitize_cloud(cloud)
    if begins_with_vv:
        split.append(cloud[:2])
        cloud = cloud[2:]
    while len(cloud) >= 3:
        split.append(cloud[:3])
        cloud = cloud[3:]
    if cloud:
        split.append(cloud)
    if len(split) == 1:
        split.append('')
    return split


def get_clouds(wxdata: [str]) -> ([str], list):
    """Returns the report list and removed list of split cloud layers"""
    clouds = []
    for i, item in reversed(list(enumerate(wxdata))):
        if item[:3] in CLOUD_LIST:
            clouds.append(split_cloud(wxdata.pop(i), False))
        elif item[:2] == 'VV':
            clouds.append(split_cloud(wxdata.pop(i), True))
    return wxdata, sorted(clouds, key=lambda pair: (pair[1], pair[0]))


def get_flight_rules(vis: str, cloud: [str]) -> int:
    """Returns int based on current flight rules from parsed METAR data
    0=VFR, 1=MVFR, 2=IFR, 3=LIFR
    Note: Common practice is to report IFR if visibility unavailable
    """
    # Parse visibility
    if not vis or is_unknown(vis):
        return 2
    elif vis == 'P6':
        vis = 10
    elif '/' in vis:
        vis = 0 if vis[0] == 'M' else int(vis.split('/')[0]) / int(vis.split('/')[1])
    # Convert meters to miles
    elif len(vis) == 4 and vis.isdigit():
        vis = int(vis) * 0.000621371
    else:
        vis = int(vis)
    # Parse ceiling
    cld = int(cloud[1]) if cloud else 99
    # Determine flight rules
    if (vis <= 5) or (cld <= 30):
        if (vis < 3) or (cld < 10):
            if (vis < 1) or (cld < 5):
                return 3  # LIFR
            return 2  # IFR
        return 1  # MVFR
    return 0  # VFR


def get_taf_flight_rules(lines: [str]) -> [str]:
    """Get flight rules by looking for missing data in prior reports"""
    for i, line in enumerate(lines):
        temp_vis, temp_cloud = line['Visibility'], line['Cloud-List']
        for report in reversed(lines[:i]):
            if is_not_tempo_or_prob(report['Type']):
                if temp_vis == '':
                    temp_vis = report['Visibility']
                if 'SKC' in report['Other-List'] or 'CLR' in report['Other-List']:
                    temp_cloud = 'tempClear'
                elif temp_cloud == []:
                    temp_cloud = report['Cloud-List']
                if temp_vis != '' and temp_cloud != []:
                    break
        if temp_cloud == 'tempClear':
            temp_cloud = []
        line['Flight-Rules'] = FLIGHT_RULES[get_flight_rules(temp_vis, get_ceiling(temp_cloud))]
    return lines


def get_ceiling(clouds: [[str]]) -> [str]:
    """Returns list of ceiling layer from Cloud-List or None if none found
    Assumes that the clouds are already sorted lowest to highest
    Only 'Broken', 'Overcast', and 'Vertical Visibility' are considdered ceilings
    Prevents errors due to lack of cloud information (eg. '' or 'FEW///')
    """
    for cloud in clouds:
        if len(cloud) > 1 and cloud[1].isdigit() and cloud[0] in ['OVC', 'BKN', 'VV']:
            return cloud
    return None


def parse_remarks(rmk: str) -> {str: str}:
    """Finds temperature and dewpoint decimal values from the remarks"""
    rmkdata = {}
    for item in rmk.split(' '):
        if len(item) in [5, 9] and item[0] == 'T' and item[1:].isdigit():
            if item[1] == '1':
                rmkdata['Temp-Decimal'] = '-' + item[2].replace('0', '') + item[3] + '.' + item[4]
            elif item[1] == '0':
                rmkdata['Temp-Decimal'] = item[2].replace('0', '') + item[3] + '.' + item[4]
            if len(item) == 9:
                if item[5] == '1':
                    rmkdata['Dew-Decimal'] = '-' + item[6].replace('0', '') + item[7] + '.' + item[8]
                elif item[5] == '0':
                    rmkdata['Dew-Decimal'] = item[6].replace('0', '') + item[7] + '.' + item[8]
    return rmkdata



"""
Michael duPont - michael@mdupont.com
AVWX-Engine : avwx/exceptions.py

Contains avwx custom exceptions
"""


class BadStation(Exception):
    """Station does not exists"""
    pass


class InvalidRequest(Exception):
    """Unable to fetch data"""
    pass



"""
Michael duPont - michael@mdupont.com
AVWX-Engine : setup.py

avwx package metadata
"""

from setuptools import setup

setup(
    name='avwx-engine',
    version='0.11',
    description='Aviation weather report parsing library',
    url='https://github.com/flyinactor91/AVWX-Engine',
    author='Michael duPont',
    author_email='michael@mdupont.com',
    license='MIT',
    packages=[
        'avwx'
    ],
    package_data={
        'avwx': [
            'stations.json'
        ]
    }
)



"""
Michael duPont - michael@mdupont.com
AVWX-Engine : avwx/translate.py

Contains functions for translating report data
"""

from avwx import core, remarks
from avwx.static import CLOUD_TRANSLATIONS, WX_TRANSLATIONS, \
                        TURBULANCE_CONDITIONS, ICING_CONDITIONS


def get_cardinal_direction(wdir: str) -> str:
    """Returns the cardinal direction (NSEW) for a degree direction
    Wind Direction - Cheat Sheet
    (360) -- 011/012 -- 033/034 -- (045) -- 056/057 -- 078/079 -- (090)
    (090) -- 101/102 -- 123/124 -- (135) -- 146/147 -- 168/169 -- (180)
    (180) -- 191/192 -- 213/214 -- (225) -- 236/237 -- 258/259 -- (270)
    (270) -- 281/282 -- 303/304 -- (315) -- 326/327 -- 348/349 -- (360)
    """
    ret = ''
    if not isinstance(wdir, int):
        wdir = int(wdir)
    if 304 <= wdir <= 360 or 0 <= wdir <= 56:
        ret += 'N'
        if 304 <= wdir <= 348:
            if 327 <= wdir <= 348:
                ret += 'N'
            ret += 'W'
        elif 11 <= wdir <= 56:
            if 11 <= wdir <= 33:
                ret += 'N'
            ret += 'E'
    elif 124 <= wdir <= 236:
        ret += 'S'
        if 124 <= wdir <= 168:
            if 147 <= wdir <= 168:
                ret += 'S'
            ret += 'E'
    elif 57 <= wdir <= 123:
        ret += 'E'
        if 57 <= wdir <= 78:
            ret += 'NE'
        elif 102 <= wdir <= 123:
            ret += 'SE'
    elif 237 <= wdir <= 303:
        ret += 'W'
        if 237 <= wdir <= 258:
            ret += 'SW'
        elif 282 <= wdir <= 303:
            ret += 'NW'
    return ret


def wind(wdir: str, wspd: str, wgst: str, wvar: [str] = None, unit: str = 'kt', cardinals: bool = True) -> str:
    """Format wind elements into a readable sentence
    Returns the translation string
    Ex: NNE-020 (variable 010 to 040) at 14kt gusting to 20kt
    """
    ret = ''
    if wdir == '000':
        ret += 'Calm'
    elif wdir.isdigit():
        if cardinals:
            ret += get_cardinal_direction(wdir) + '-'
        ret += wdir
    elif wdir == 'VRB':
        ret += 'Variable'
    else:
        ret += wdir
    if wvar and isinstance(wvar, list):
        ret += ' (variable {low} to {high})'.format(low=wvar[0], high=wvar[1])
    if wspd and wspd not in ('0', '00'):
        ret += ' at {speed}{unit}'.format(speed=wspd, unit=unit)
    if wgst:
        ret += ' gusting to {speed}{unit}'.format(speed=wgst, unit=unit)
    return ret


def visibility(vis: str, unit: str = 'm') -> str:
    """Formats a visibility element into a string with both km and sm values
    Ex: 8km ( 5sm )
    """
    if vis == 'P6':
        return 'Greater than 6sm ( >9999m )'
    if vis == 'M1/4':
        return 'Less than .25sm ( <0400m )'
    if '/' in vis and not core.is_unknown(vis):
        vis = float(vis[:vis.find('/')]) / int(vis[vis.find('/') + 1:])
    try:
        float(vis)
    except ValueError:
        return ''
    if unit == 'm':
        converted = float(vis) * 0.000621371
        converted = str(round(converted, 1)).replace('.0', '') + 'sm'
        vis = str(round(int(vis) / 1000, 1)).replace('.0', '')
        unit = 'km'
    elif unit == 'sm':
        converted = float(vis) / 0.621371
        converted = str(round(converted, 1)).replace('.0', '') + 'km'
        vis = str(vis).replace('.0', '')
    else:
        return ''
    return vis + unit + ' (' + converted + ')'


def temperature(temp: str, unit: str = 'C') -> str:
    """Formats a temperature element into a string with both C and F values
    Used for both Temp and Dew
    Ex: 34C (93F)
    """
    temp = temp.replace('M', '-')
    try:
        int(temp)
    except ValueError:
        return ''
    unit = unit.upper()
    if unit == 'C':
        converted = int(temp) * 1.8 + 32
        converted = str(int(round(converted))) + 'F'
    elif unit == 'F':
        converted = (int(temp) - 32) / 1.8
        converted = str(int(round(converted))) + 'C'
    else:
        return ''
    return temp + unit + ' (' + converted + ')'


def altimeter(alt: str, unit: str = 'hPa') -> str:
    """Formats the altimter element into a string with hPa and inHg values
    Ex: 30.11 inHg (10.20 hPa)
    """
    if not alt.isdigit():
        if len(alt) == 5 and alt[1:].isdigit():
            alt = alt[1:]
        else:
            return ''
    if unit == 'hPa':
        converted = float(alt) / 33.8638866667
        converted = str(round(converted, 2)) + 'inHg'
    elif unit == 'inHg':
        alt = alt[:2] + '.' + alt[2:]
        converted = float(alt) * 33.8638866667
        converted = str(int(round(converted))) + 'hPa'
    else:
        return ''
    return alt + unit + ' (' + converted + ')'


def clouds(clds: [str], unit: str = 'ft') -> str:
    """Format cloud list into a readable sentence
    Returns the translation string
    Ex: Broken layer at 2200ft (Cumulonimbus), Overcast layer at 3600ft - Reported AGL
    """
    ret = []
    for cloud in clds:
        if len(cloud) == 2 and cloud[1].isdigit() and cloud[0] in CLOUD_TRANSLATIONS:
            ret.append(CLOUD_TRANSLATIONS[cloud[0]].format(int(cloud[1]) * 100, unit))
        elif len(cloud) == 3 and cloud[1].isdigit() \
            and cloud[0] in CLOUD_TRANSLATIONS and cloud[2] in CLOUD_TRANSLATIONS:
            cloudstr = CLOUD_TRANSLATIONS[cloud[0]] + ' (' + CLOUD_TRANSLATIONS[cloud[2]] + ')'
            ret.append(cloudstr.format(int(cloud[1]) * 100, unit))
    if ret:
        return ', '.join(ret) + ' - Reported AGL'
    return 'Sky clear'


def wxcode(wxstr: str) -> str:
    """Translates weather codes into readable strings
    Returns translated string of variable length
    """
    if wxstr[0] == '+':
        ret = 'Heavy '
        wxstr = wxstr[1:]
    elif wxstr[0] == '-':
        ret = 'Light '
        wxstr = wxstr[1:]
    else:
        ret = ''
    #Return wxstr if wxstr is not a code, ex R03/03002V03
    if len(wxstr) not in [2, 4, 6]:
        return wxstr
    for _ in range(len(wxstr) // 2):
        if wxstr[:2] in WX_TRANSLATIONS:
            ret += WX_TRANSLATIONS[wxstr[:2]] + ' '
        else:
            ret += wxstr[:2]
        wxstr = wxstr[2:]
    return ret.strip(' ')


def other_list(wxcodes: [str]) -> str:
    """Translate the list of wx codes (otherList) into a readable sentence
    Returns the translation string
    """
    ret = []
    for item in wxcodes:
        ret.append(wxcode(item))
    return ', '.join(ret)


def wind_shear(shear: str, unit_alt: str = 'ft', unit_wnd: str = 'kt') -> str:
    """Translate wind shear into a readable string
    Ex: Wind shear 2000ft from 140 at 30kt
    """
    if not shear or 'WS' not in shear or '/' not in shear:
        return ''
    shear = shear[2:].split('/')
    return 'Wind shear {alt}{unit_alt} from {winddir} at {speed}{unit_wind}'.format(
        alt=int(shear[0]) * 100, unit_alt=unit_alt, winddir=shear[1][:3],
        speed=shear[1][3:], unit_wind=unit_wnd)


def turb_ice(turbice: [str], unit: str = 'ft') -> str:
    """Translate the list of turbulance or icing into a readable sentence
    Ex: Occasional moderate turbulence in clouds from 3000ft to 14000ft
    """
    if not turbice:
        return ''
    #Determine turbulance or icing
    if turbice[0][0] == '5':
        conditions = TURBULANCE_CONDITIONS
    elif turbice[0][0] == '6':
        conditions = ICING_CONDITIONS
    else:
        return ''
    #Create list of split items (type, floor, height)
    split = []
    for item in turbice:
        if len(item) == 6:
            split.append([item[1:2], item[2:5], item[5:6]])
    #Combine items that cover a layer greater than 9000ft
    for i in reversed(range(len(split) - 1)):
        if split[i][2] == '9' and split[i][0] == split[i + 1][0] \
            and int(split[i + 1][1]) == (int(split[i][1]) + int(split[i][2]) * 10):
            split[i][2] = str(int(split[i][2]) + int(split[i + 1][2]))
            split.pop(i + 1)
    #Return joined, formatted string from split items
    return ', '.join(['{conditions} from {low_alt}{unit} to {high_alt}{unit}'.format(
        conditions=conditions[item[0]], low_alt=int(item[1]) * 100,
        high_alt=int(item[1]) * 100 + int(item[2]) * 1000, unit=unit) for item in split])


def min_max_temp(temp: str, unit: str = 'C') -> str:
    """Format the Min and Max temp elemets into a readable string
    Ex: Maximum temperature of 23C (73F) at 18-15:00Z
    """
    if not temp or len(temp) < 7:
        return ''
    if temp[:2] == 'TX':
        temp_type = 'Maximum'
    elif temp[:2] == 'TN':
        temp_type = 'Minimum'
    else:
        return ''
    temp = temp[2:].replace('M', '-').replace('Z', '').split('/')
    if len(temp[1]) > 2:
        temp[1] = temp[1][:2] + '-' + temp[1][2:]
    return '{temp_type} temperature of {temp} at {time}:00Z'.format(
        temp_type=temp_type, temp=temperature(temp[0], unit), time=temp[1])


def shared(wxdata: [str], units: {str: str}) -> {str: str}:
    """Translate Visibility, Altimeter, Clouds, and Other"""
    translations = {}
    translations['Visibility'] = visibility(wxdata['Visibility'], units['Visibility'])
    translations['Altimeter'] = altimeter(wxdata['Altimeter'], units['Altimeter'])
    translations['Clouds'] = clouds(wxdata['Cloud-List'], units['Altitude'])
    translations['Other'] = other_list(wxdata['Other-List'])
    return translations


def metar(wxdata: {str: object}) -> {str: str}:
    """Translate the results of metar.parse
    Keys: Wind, Visibility, Clouds, Temperature, Dewpoint, Altimeter, Other
    """
    units = wxdata['Units']
    translations = shared(wxdata, units)
    translations['Wind'] = wind(wxdata['Wind-Direction'], wxdata['Wind-Speed'],
                                wxdata['Wind-Gust'], wxdata['Wind-Variable-Dir'],
                                units['Wind-Speed'])
    translations['Temperature'] = temperature(wxdata['Temperature'], units['Temperature'])
    translations['Dewpoint'] = temperature(wxdata['Dewpoint'], units['Temperature'])
    translations['Remarks'] = remarks.translate(wxdata['Remarks'])
    return translations


def taf(wxdata: {str: object}) -> {str: str}:
    """Translate the results of taf.parse
    Keys: Forecast, Min-Temp, Max-Temp
    Forecast keys: Wind, Visibility, Clouds, Altimeter, Wind-Shear, Turbulance, Icing, Other
    """
    translations = {'Forecast': []}
    units = wxdata['Units']
    for line in wxdata['Forecast']:
        trans = shared(line, units)
        trans['Wind'] = wind(line['Wind-Direction'], line['Wind-Speed'],
                             line['Wind-Gust'], unit=units['Wind-Speed'])
        trans['Wind-Shear'] = wind_shear(line['Wind-Shear'],
                                         wxdata['Units']['Altitude'],
                                         units['Wind-Speed'])
        trans['Turbulance'] = turb_ice(line['Turb-List'], units['Altitude'])
        trans['Icing'] = turb_ice(line['Icing-List'], units['Altitude'])
        # Remove false 'Sky Clear' if line type is 'BECMG'
        if line['Type'] == 'BECMG' and trans['Clouds'] == 'Sky clear':
            trans['Clouds'] = ''
        translations['Forecast'].append(trans)
    translations['Min-Temp'] = min_max_temp(wxdata['Min-Temp'], units['Temperature'])
    translations['Max-Temp'] = min_max_temp(wxdata['Max-Temp'], units['Temperature'])
    translations['Remarks'] = remarks.translate(wxdata['Remarks'])
    return translations



"""
Michael duPont - michael@mdupont.com
AVWX-Engine : avwx/metar.py

Contains METAR-specific functions for fetching and parsing
"""

# stdlib
from copy import copy
# module
from avwx import core, service
from avwx.static import NA_UNITS, IN_UNITS, FLIGHT_RULES

def fetch(station: str) -> str:
    """Get METAR report for 'station' from www.aviationweather.gov
    Returns METAR report string or raises an error
    Maintains backwards compatability but uses the new AddsRequest object
    """
    return service.get_service(station)('metar').fetch(station)


def parse(station: str, txt: str) -> {str: object}:
    """Returns a dictionary of parsed METAR data
    Keys: Station, Time, Wind-Direction, Wind-Speed, Wind-Gust, Wind-Variable-Dir,
          Visibility, Runway-Vis-List, Altimeter, Temperature, Dewpoint,
          Cloud-List, Other-List, Remarks, Raw-Report, Units
    Units is dict of identified units of measurement for each field
    """
    core.valid_station(station)
    return parse_na(txt) if core.uses_na_format(station[:2]) else parse_in(txt)


def parse_na(txt: str) -> {str: object}:
    """Parser for the North American METAR variant"""
    units = copy(NA_UNITS)
    wxresp = {'Raw-Report': txt}
    txt = core.sanitize_report_string(txt)
    wxdata, wxresp['Remarks'] = core.get_remarks(txt)
    wxdata, wxresp['Runway-Vis-List'], _ = core.sanitize_report_list(wxdata)
    wxdata, wxresp['Station'], wxresp['Time'] = core.get_station_and_time(wxdata)
    wxdata, wxresp['Cloud-List'] = core.get_clouds(wxdata)
    wxdata, units, wxresp['Wind-Direction'], wxresp['Wind-Speed'], \
        wxresp['Wind-Gust'], wxresp['Wind-Variable-Dir'] = core.get_wind(wxdata, units)
    wxdata, units, wxresp['Altimeter'] = core.get_altimeter(wxdata, units, 'NA')
    wxdata, units, wxresp['Visibility'] = core.get_visibility(wxdata, units)
    wxresp['Other-List'], wxresp['Temperature'], wxresp['Dewpoint'] = core.get_temp_and_dew(wxdata)
    wxresp['Units'] = units
    condition = core.get_flight_rules(wxresp['Visibility'], core.get_ceiling(wxresp['Cloud-List']))
    wxresp['Flight-Rules'] = FLIGHT_RULES[condition]
    wxresp['Remarks-Info'] = core.parse_remarks(wxresp['Remarks'])
    return wxresp


def parse_in(txt: str) -> {str: object}:
    """Parser for the International METAR variant"""
    units = copy(IN_UNITS)
    wxresp = {'Raw-Report': txt}
    txt = core.sanitize_report_string(txt)
    wxdata, wxresp['Remarks'] = core.get_remarks(txt)
    wxdata, wxresp['Runway-Vis-List'], _ = core.sanitize_report_list(wxdata)
    wxdata, wxresp['Station'], wxresp['Time'] = core.get_station_and_time(wxdata)
    if 'CAVOK' not in wxdata:
        wxdata, wxresp['Cloud-List'] = core.get_clouds(wxdata)
    wxdata, units, wxresp['Wind-Direction'], wxresp['Wind-Speed'], \
        wxresp['Wind-Gust'], wxresp['Wind-Variable-Dir'] = core.get_wind(wxdata, units)
    wxdata, units, wxresp['Altimeter'] = core.get_altimeter(wxdata, units, 'IN')
    if 'CAVOK' in wxdata:
        wxresp['Visibility'] = '9999'
        wxresp['Cloud-List'] = []
        wxdata.remove('CAVOK')
    else:
        wxdata, units, wxresp['Visibility'] = core.get_visibility(wxdata, units)
    wxresp['Other-List'], wxresp['Temperature'], wxresp['Dewpoint'] = core.get_temp_and_dew(wxdata)
    wxresp['Units'] = units
    condition = core.get_flight_rules(wxresp['Visibility'], core.get_ceiling(wxresp['Cloud-List']))
    wxresp['Flight-Rules'] = FLIGHT_RULES[condition]
    wxresp['Remarks-Info'] = core.parse_remarks(wxresp['Remarks'])
    return wxresp
